import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface

class JSONSchemaGenerator(GeneratorInterface):
    """
    JSONSchemaGenerator is a class that implements the GeneratorInterface and is responsible for generating
    JSONSchema code based on B-UML models.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """
        
    def __init__(self, model: DomainModel, output_dir: str = None):
        super().__init__(model, output_dir)
        # Add enums to TYPES dictionary

    def generate(self):
        """
        Generates JSONSchema code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file named json_schema.json
        """
        file_path = self.build_generation_path(file_name="json_schema.json")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template('json_schema.json.j2')
        with open(file_path, mode="w") as f:
            generated_code = template.render(
                classes=self.model.classes_sorted_by_inheritance(),
                enumerations = self.model.get_enumerations(),
            )
            f.write(generated_code)
            print("Code generated in the location: " + file_path)



# Generated from OD.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,17,80,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,2,10,7,10,2,11,7,11,1,0,1,0,1,0,4,0,28,
        8,0,11,0,12,0,29,1,0,1,0,1,1,1,1,1,1,1,1,3,1,38,8,1,1,1,3,1,41,8,
        1,1,2,1,2,1,3,1,3,1,4,1,4,5,4,49,8,4,10,4,12,4,52,9,4,1,4,1,4,1,
        5,1,5,1,5,1,5,1,6,1,6,1,7,4,7,63,8,7,11,7,12,7,64,1,8,1,8,1,8,1,
        8,1,8,3,8,72,8,8,1,9,1,9,1,10,1,10,1,11,1,11,1,11,0,0,12,0,2,4,6,
        8,10,12,14,16,18,20,22,0,3,2,0,13,15,17,17,1,0,5,10,1,0,13,14,74,
        0,24,1,0,0,0,2,33,1,0,0,0,4,42,1,0,0,0,6,44,1,0,0,0,8,46,1,0,0,0,
        10,55,1,0,0,0,12,59,1,0,0,0,14,62,1,0,0,0,16,66,1,0,0,0,18,73,1,
        0,0,0,20,75,1,0,0,0,22,77,1,0,0,0,24,27,5,11,0,0,25,28,3,2,1,0,26,
        28,3,16,8,0,27,25,1,0,0,0,27,26,1,0,0,0,28,29,1,0,0,0,29,27,1,0,
        0,0,29,30,1,0,0,0,30,31,1,0,0,0,31,32,5,12,0,0,32,1,1,0,0,0,33,34,
        5,1,0,0,34,37,3,4,2,0,35,36,5,2,0,0,36,38,3,6,3,0,37,35,1,0,0,0,
        37,38,1,0,0,0,38,40,1,0,0,0,39,41,3,8,4,0,40,39,1,0,0,0,40,41,1,
        0,0,0,41,3,1,0,0,0,42,43,5,13,0,0,43,5,1,0,0,0,44,45,5,13,0,0,45,
        7,1,0,0,0,46,50,5,3,0,0,47,49,3,10,5,0,48,47,1,0,0,0,49,52,1,0,0,
        0,50,48,1,0,0,0,50,51,1,0,0,0,51,53,1,0,0,0,52,50,1,0,0,0,53,54,
        5,4,0,0,54,9,1,0,0,0,55,56,3,12,6,0,56,57,5,2,0,0,57,58,3,14,7,0,
        58,11,1,0,0,0,59,60,5,13,0,0,60,13,1,0,0,0,61,63,7,0,0,0,62,61,1,
        0,0,0,63,64,1,0,0,0,64,62,1,0,0,0,64,65,1,0,0,0,65,15,1,0,0,0,66,
        67,3,18,9,0,67,68,3,20,10,0,68,71,3,18,9,0,69,70,5,2,0,0,70,72,3,
        22,11,0,71,69,1,0,0,0,71,72,1,0,0,0,72,17,1,0,0,0,73,74,5,13,0,0,
        74,19,1,0,0,0,75,76,7,1,0,0,76,21,1,0,0,0,77,78,7,2,0,0,78,23,1,
        0,0,0,7,27,29,37,40,50,64,71
    ]

class ODParser ( Parser ):

    grammarFileName = "OD.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'Object'", "':'", "'{'", "'}'", "'<|--'", 
                     "'o--'", "'--'", "'..'", "'-->'", "'..>'", "'@startuml'", 
                     "'@enduml'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "Start", "End", 
                      "IDENTIFIER", "STRING", "NUMBER", "WHITESPACE", "DATE" ]

    RULE_objectDiagram = 0
    RULE_objectDeclaration = 1
    RULE_objectName = 2
    RULE_className = 3
    RULE_propertiesBlock = 4
    RULE_property = 5
    RULE_propertyName = 6
    RULE_propertyValue = 7
    RULE_linkDeclaration = 8
    RULE_linkObjectName = 9
    RULE_linkType = 10
    RULE_linkName = 11

    ruleNames =  [ "objectDiagram", "objectDeclaration", "objectName", "className", 
                   "propertiesBlock", "property", "propertyName", "propertyValue", 
                   "linkDeclaration", "linkObjectName", "linkType", "linkName" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    Start=11
    End=12
    IDENTIFIER=13
    STRING=14
    NUMBER=15
    WHITESPACE=16
    DATE=17

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ObjectDiagramContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def Start(self):
            return self.getToken(ODParser.Start, 0)

        def End(self):
            return self.getToken(ODParser.End, 0)

        def objectDeclaration(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ODParser.ObjectDeclarationContext)
            else:
                return self.getTypedRuleContext(ODParser.ObjectDeclarationContext,i)


        def linkDeclaration(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ODParser.LinkDeclarationContext)
            else:
                return self.getTypedRuleContext(ODParser.LinkDeclarationContext,i)


        def getRuleIndex(self):
            return ODParser.RULE_objectDiagram

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObjectDiagram" ):
                listener.enterObjectDiagram(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObjectDiagram" ):
                listener.exitObjectDiagram(self)




    def objectDiagram(self):

        localctx = ODParser.ObjectDiagramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_objectDiagram)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 24
            self.match(ODParser.Start)
            self.state = 27 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 27
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [1]:
                    self.state = 25
                    self.objectDeclaration()
                    pass
                elif token in [13]:
                    self.state = 26
                    self.linkDeclaration()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 29 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==1 or _la==13):
                    break

            self.state = 31
            self.match(ODParser.End)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ObjectDeclarationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def objectName(self):
            return self.getTypedRuleContext(ODParser.ObjectNameContext,0)


        def className(self):
            return self.getTypedRuleContext(ODParser.ClassNameContext,0)


        def propertiesBlock(self):
            return self.getTypedRuleContext(ODParser.PropertiesBlockContext,0)


        def getRuleIndex(self):
            return ODParser.RULE_objectDeclaration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObjectDeclaration" ):
                listener.enterObjectDeclaration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObjectDeclaration" ):
                listener.exitObjectDeclaration(self)




    def objectDeclaration(self):

        localctx = ODParser.ObjectDeclarationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_objectDeclaration)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 33
            self.match(ODParser.T__0)
            self.state = 34
            self.objectName()
            self.state = 37
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==2:
                self.state = 35
                self.match(ODParser.T__1)
                self.state = 36
                self.className()


            self.state = 40
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==3:
                self.state = 39
                self.propertiesBlock()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ObjectNameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(ODParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return ODParser.RULE_objectName

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObjectName" ):
                listener.enterObjectName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObjectName" ):
                listener.exitObjectName(self)




    def objectName(self):

        localctx = ODParser.ObjectNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_objectName)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 42
            self.match(ODParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ClassNameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(ODParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return ODParser.RULE_className

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterClassName" ):
                listener.enterClassName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitClassName" ):
                listener.exitClassName(self)




    def className(self):

        localctx = ODParser.ClassNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_className)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 44
            self.match(ODParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropertiesBlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def property_(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ODParser.PropertyContext)
            else:
                return self.getTypedRuleContext(ODParser.PropertyContext,i)


        def getRuleIndex(self):
            return ODParser.RULE_propertiesBlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropertiesBlock" ):
                listener.enterPropertiesBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropertiesBlock" ):
                listener.exitPropertiesBlock(self)




    def propertiesBlock(self):

        localctx = ODParser.PropertiesBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_propertiesBlock)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 46
            self.match(ODParser.T__2)
            self.state = 50
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==13:
                self.state = 47
                self.property_()
                self.state = 52
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 53
            self.match(ODParser.T__3)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropertyContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def propertyName(self):
            return self.getTypedRuleContext(ODParser.PropertyNameContext,0)


        def propertyValue(self):
            return self.getTypedRuleContext(ODParser.PropertyValueContext,0)


        def getRuleIndex(self):
            return ODParser.RULE_property

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterProperty" ):
                listener.enterProperty(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitProperty" ):
                listener.exitProperty(self)




    def property_(self):

        localctx = ODParser.PropertyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_property)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 55
            self.propertyName()
            self.state = 56
            self.match(ODParser.T__1)
            self.state = 57
            self.propertyValue()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropertyNameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(ODParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return ODParser.RULE_propertyName

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropertyName" ):
                listener.enterPropertyName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropertyName" ):
                listener.exitPropertyName(self)




    def propertyName(self):

        localctx = ODParser.PropertyNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_propertyName)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 59
            self.match(ODParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropertyValueContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self, i:int=None):
            if i is None:
                return self.getTokens(ODParser.IDENTIFIER)
            else:
                return self.getToken(ODParser.IDENTIFIER, i)

        def STRING(self, i:int=None):
            if i is None:
                return self.getTokens(ODParser.STRING)
            else:
                return self.getToken(ODParser.STRING, i)

        def NUMBER(self, i:int=None):
            if i is None:
                return self.getTokens(ODParser.NUMBER)
            else:
                return self.getToken(ODParser.NUMBER, i)

        def DATE(self, i:int=None):
            if i is None:
                return self.getTokens(ODParser.DATE)
            else:
                return self.getToken(ODParser.DATE, i)

        def getRuleIndex(self):
            return ODParser.RULE_propertyValue

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropertyValue" ):
                listener.enterPropertyValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropertyValue" ):
                listener.exitPropertyValue(self)




    def propertyValue(self):

        localctx = ODParser.PropertyValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_propertyValue)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 62 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 61
                    _la = self._input.LA(1)
                    if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 188416) != 0)):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()

                else:
                    raise NoViableAltException(self)
                self.state = 64 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,5,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LinkDeclarationContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def linkObjectName(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ODParser.LinkObjectNameContext)
            else:
                return self.getTypedRuleContext(ODParser.LinkObjectNameContext,i)


        def linkType(self):
            return self.getTypedRuleContext(ODParser.LinkTypeContext,0)


        def linkName(self):
            return self.getTypedRuleContext(ODParser.LinkNameContext,0)


        def getRuleIndex(self):
            return ODParser.RULE_linkDeclaration

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLinkDeclaration" ):
                listener.enterLinkDeclaration(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLinkDeclaration" ):
                listener.exitLinkDeclaration(self)




    def linkDeclaration(self):

        localctx = ODParser.LinkDeclarationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_linkDeclaration)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 66
            self.linkObjectName()
            self.state = 67
            self.linkType()
            self.state = 68
            self.linkObjectName()
            self.state = 71
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==2:
                self.state = 69
                self.match(ODParser.T__1)
                self.state = 70
                self.linkName()


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LinkObjectNameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(ODParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return ODParser.RULE_linkObjectName

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLinkObjectName" ):
                listener.enterLinkObjectName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLinkObjectName" ):
                listener.exitLinkObjectName(self)




    def linkObjectName(self):

        localctx = ODParser.LinkObjectNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_linkObjectName)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 73
            self.match(ODParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LinkTypeContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ODParser.RULE_linkType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLinkType" ):
                listener.enterLinkType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLinkType" ):
                listener.exitLinkType(self)




    def linkType(self):

        localctx = ODParser.LinkTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_linkType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 75
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & 2016) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LinkNameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(ODParser.STRING, 0)

        def IDENTIFIER(self):
            return self.getToken(ODParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return ODParser.RULE_linkName

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLinkName" ):
                listener.enterLinkName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLinkName" ):
                listener.exitLinkName(self)




    def linkName(self):

        localctx = ODParser.LinkNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_linkName)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 77
            _la = self._input.LA(1)
            if not(_la==13 or _la==14):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import pytest

from besser.BUML.metamodel.structural import *
from besser.BUML.metamodel.gui import *


def test_named_element():
    named_element: NamedElement = NamedElement(name="element1")
    assert named_element.name == "element1"

# Test: Do not have two modules with the same name in an application.
def test_unique_module_names():
    module1: Module = Module(name="module1", screens=[])
    module2: Module = Module(name="module2", screens=[])
    my_app: Application = Application(name="application1", package="", versionCode="", versionName="", description="", screenCompatibility=False, modules=[module1, module2])
    with pytest.raises(ValueError) as excinfo:
        # Try to create a module with the same name as module1
        module_duplicate: Module = Module(name="module1", screens=[])
        my_app.modules = {module1, module2, module_duplicate}
    assert "An app cannot have two modules with the same name" in str(excinfo.value)

# Test: Do not have two screens with the same name in an application.
def test_unique_screen_names():
    screen1: Screen = Screen(name="screen1", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={})
    screen2: Screen = Screen(name="screen2", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={})
    module1: Module = Module(name="module1", screens={screen1, screen2})
    my_app: Application = Application(name="application1", package="", versionCode="", versionName="", description="", screenCompatibility=False, modules={module1})
    with pytest.raises(ValueError) as excinfo:
        # Try to create a screen with the same name as screen1
        screen_duplicate: Screen = Screen(name="screen1", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={})
        module1.screens = {screen1, screen2, screen_duplicate}
    assert "A module cannot have two screens with the same name" in str(excinfo.value)

# Test: Do not have two items with the same name in a list.
def test_unique_item_names():
    item1: ModelElement = ModelElement(name="item1", dataSourceClass="", fields={})
    item2: ModelElement = ModelElement(name="item2", dataSourceClass="", fields={})
    list1: DataList = DataList(name="list1", description="", list_sources={item1, item2})
    with pytest.raises(ValueError) as excinfo:
        # Try to create a item with the same name as item1
        item_duplicate: ModelElement = ModelElement(name="item1", dataSourceClass="", fields={})
        list1.list_sources={item1, item2, item_duplicate}
    assert "A list cannot have two items with the same name" in str(excinfo.value)

def test_button_buttonType_must_be_defined():
    button: Button = Button(name="button", label="View List", description="", buttonType="", actionType=ButtonActionType.Add)
    
    with pytest.raises(ValueError) as excinfo:
        if button.buttonType == "":
            raise ValueError("buttonType must be defined")
        screen1: Screen = Screen(name="screen1", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={button})
    assert "buttonType must be defined" in str(excinfo.value)

def test_button_actionType_must_be_defined():
    button: Button = Button(name="button", label="View List", description="", buttonType=ButtonType.FloatingActionButton, actionType="")
    
    with pytest.raises(ValueError) as excinfo:
        if button.actionType == "":
            raise ValueError("actionType must be defined")
        screen1: Screen = Screen(name="screen1", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={button})
    assert "actionType must be defined" in str(excinfo.value)

def test_button_properties_must_be_defined():
    button: Button = Button(name="button", label="View List", description="", buttonType="", actionType="")
    
    with pytest.raises(ValueError) as excinfo:
        if button.buttonType == "" and button.actionType== "":
            raise ValueError("buttonType and actionType must be defined")
        screen1: Screen = Screen(name="screen1", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={button})
    assert "buttonType and actionType must be defined" in str(excinfo.value)


def test_list_sources_must_be_defined():
    datasource: ModelElement = ModelElement(name="Data Source", dataSourceClass="", fields=[])
    myList: DataList = DataList(name="MyList", description="A diverse group of elements", list_sources={})

    with pytest.raises(ValueError) as excinfo:
        if len(myList.list_sources) == 0:
            raise ValueError("list_sources must be defined")

        screen1: Screen = Screen(name="screen1", description="", x_dpi="", y_dpi="", size="SmallScreen", view_elements={myList})

    assert "list_sources must be defined" in str(excinfo.value)

def test_fields_synchronized_attributes():
    # Class1 attributes definition
    class1_name: Property = Property(name="name", type="int")
    class1: Class = Class (name="Class1", attributes=[class1_name])
    # Class2 attributes definition
    class2_name: Property = Property(name="name", type="int")
    class1: Class = Class (name="Class2", attributes=[class2_name])
    #class1_DataSource definition
    datasource_class1: ModelElement = ModelElement(name="Class1 Data Source", dataSourceClass=class1, fields=[class2_name])
    with pytest.raises(ValueError) as excinfo:
        if datasource_class1.fields not in datasource_class1.dataSourceClass.attributes:
            raise ValueError("Fields must be synchronized with the dataSourceClass attributes")
        class1_List: DataList=DataList(name="Class1 List", description="A diverse group of elements", list_sources={datasource_class1})
   
    assert "Fields must be synchronized with the dataSourceClass attributes" in str(excinfo.value)

def test_associations():
    # Class1 attributes definition
    class1_name: Property = Property(name="name", type="int")
    class1: Class = Class(name="Class1", attributes=[class1_name])
    # Class2 attributes definition
    class2_name: Property = Property(name="name", type="int")
    class2: Class = Class(name="Class2", attributes=[class2_name])
    # Class1-Class2 association definition
    end1: Property = Property(name="end1", type=class2)
    end2: Property = Property(name="end2", type=class1)
    class1_class2_association: BinaryAssociation = BinaryAssociation(name="class1_class2_association", ends={end1, end2})
    # Domain model definition
    model: DomainModel = DomainModel(name="model", types={class1, class2}, associations={class1_class2_association})

    with pytest.raises(ValueError) as excinfo:
        if len(model.associations) != 0:
            for classConcept in model.types:
                for assoc in model.associations:
                    for end in assoc.ends:
                        if end.type.name != classConcept.name:
                            raise ValueError("All associations related to classes in 'types' must be considered in 'associations'")
        else:
            raise ValueError("All associations related to classes in 'types' must be considered in 'associations'")

    assert "All associations related to classes in 'types' must be considered in 'associations'" in str(excinfo.value)



import os
from besser.BUML.metamodel.structural.structural import DomainModel
from besser.utilities import sort_by_timestamp as sort

PRIMITIVE_TYPE_MAPPING = {
    'str': 'StringType',
    'string': 'StringType', 
    'int': 'IntegerType',
    'integer': 'IntegerType',
    'float': 'FloatType',
    'bool': 'BooleanType',
    'boolean': 'BooleanType',
    'time': 'TimeType',
    'date': 'DateType',
    'datetime': 'DateTimeType',
    'timedelta': 'TimeDeltaType',
    'any': 'AnyType'
}

# Reserved names that need special handling
RESERVED_NAMES = ['Class', 'Property', 'Method', 'Parameter', 'Enumeration']

def safe_class_name(name):
    """
    Add a suffix to class names that match reserved keywords or BUML metaclass names.
    If the name already ends with an underscore and would conflict with a reserved name,
    add a numeric suffix instead.
    
    Parameters:
    name (str): The original class name
    
    Returns:
    str: A safe variable name for the class
    """
    if not name:
        return "unnamed_class"
        
    if name.endswith('_'):
        base_name = name[:-1]
        if base_name in RESERVED_NAMES:
            return f"{name}var"
        return name
    elif name in RESERVED_NAMES:
        return f"{name}_"
    else:
        return name

def domain_model_to_code(model: DomainModel, file_path: str):
    """
    Generates Python code for a B-UML model and writes it to a specified file.

    Parameters:
    model (DomainModel): The B-UML model object containing classes, enumerations, 
        associations, and generalizations.
    file_path (str): The path where the generated code will be saved.

    Outputs:
    - A Python file containing the base code representation of the B-UML domain model.
    """
    output_dir = os.path.dirname(file_path)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not file_path.endswith('.py'):
        file_path += '.py'

    with open(file_path, 'w', encoding='utf-8') as f:
        # Write imports
        f.write("# Generated B-UML Model\n")
        f.write("from besser.BUML.metamodel.structural import (\n")
        f.write("    Class, Property, Method, Parameter,\n")
        f.write("    BinaryAssociation, Generalization, DomainModel,\n")
        f.write("    Enumeration, EnumerationLiteral, Multiplicity,\n")
        f.write("    StringType, IntegerType, FloatType, BooleanType,\n")
        f.write("    TimeType, DateType, DateTimeType, TimeDeltaType,\n")
        f.write("    AnyType, Constraint\n")
        f.write(")\n\n")

        # Write enumerations
        f.write("# Enumerations\n")
        for enum in sort(model.get_enumerations()):
            enum_var_name = safe_class_name(enum.name)
            f.write(f"{enum_var_name}: Enumeration = Enumeration(\n")
            f.write(f"    name=\"{enum.name}\",\n")
            literals_str = ",\n\t\t\t".join([f"EnumerationLiteral(name=\"{lit.name}\")" for lit in sort(enum.literals)])
            f.write(
                f"    literals={{\n"
                f"            {literals_str}"
                f"\n    }}")
            f.write("\n)\n\n")

        # Write classes
        f.write("# Classes\n")
        for cls in sort(model.get_classes()):
            cls_var_name = safe_class_name(cls.name)
            f.write(f"{cls_var_name} = Class(name=\"{cls.name}\"{', is_abstract=True' if cls.is_abstract else ''})\n")
        f.write("\n")

        # Write class members
        for cls in sort(model.get_classes()):
            cls_var_name = safe_class_name(cls.name)
            f.write(f"# {cls.name} class attributes and methods\n")

            # Write attributes
            for attr in sort(cls.attributes):
                attr_type = PRIMITIVE_TYPE_MAPPING.get(attr.type.name, safe_class_name(attr.type.name))
                visibility_str = f', visibility="{attr.visibility}"' if attr.visibility != "public" else ""
                f.write(f"{cls_var_name}_{attr.name}: Property = Property(name=\"{attr.name}\", "
                       f"type={attr_type}{visibility_str})\n")

            # Write methods
            for method in sort(cls.methods):
                method_type = PRIMITIVE_TYPE_MAPPING.get(method.type.name, safe_class_name(method.type.name)) if method.type else None
                visibility_str = f', visibility="{method.visibility}"' if method.visibility != "public" else ""
                
                # Build parameters dictionary
                params = {}
                if sort(method.parameters):
                    for param in method.parameters:
                        param_type = PRIMITIVE_TYPE_MAPPING.get(param.type.name, safe_class_name(param.type.name))
                        default_str = f", default_value='{param.default_value}'" if hasattr(param, 'default_value') and param.default_value is not None else ""
                        params[param.name] = f"Parameter(name='{param.name}', type={param_type}{default_str})"
                
                params_str = "{" + ", ".join(f"{param}" for name, param in params.items()) + "}"

                if method_type:
                    f.write(f"{cls_var_name}_m_{method.name}: Method = Method(name=\"{method.name}\""
                           f"{visibility_str}, parameters={params_str}, type={method_type})\n")
                else:
                    f.write(f"{cls_var_name}_m_{method.name}: Method = Method(name=\"{method.name}\""
                           f"{visibility_str}, parameters={params_str})\n")

            # Write assignments
            if sort(cls.attributes):
                attrs_str = ", ".join([f"{cls_var_name}_{attr.name}" for attr in cls.attributes])
                f.write(f"{cls_var_name}.attributes={{{attrs_str}}}\n")
            if sort(cls.methods):
                methods_str = ", ".join([f"{cls_var_name}_m_{method.name}" for method in cls.methods])
                f.write(f"{cls_var_name}.methods={{{methods_str}}}\n")
            f.write("\n")

        # Write relationships
        if model.associations:
            f.write("# Relationships\n")
            for assoc in sort(model.associations):
                ends_str = []
                for end in sort(assoc.ends):
                    # Determine max value for multiplicity
                    max_value = '"*"' if end.multiplicity.max == "*" else end.multiplicity.max
                    # Use safe class name for the type reference
                    type_var_name = safe_class_name(end.type.name)
                    # Build each property string with all attributes on the same line
                    end_str = (f"Property(name=\"{end.name}\", type={type_var_name}, "
                             f"multiplicity=Multiplicity({end.multiplicity.min}, {max_value})"
                             f"{', is_navigable=' + str(end.is_navigable) if end.is_navigable is not True else ''}"
                             f"{', is_composite=True' if end.is_composite is True else ''})")
                    ends_str.append(end_str)

                # Write the BinaryAssociation with each property on a new line
                f.write(
                    f"{assoc.name}: BinaryAssociation = BinaryAssociation(\n"
                    f"    name=\"{assoc.name}\",\n"
                    "    ends={\n        " + 
                    ",\n        ".join(ends_str) +
                    "\n    }\n"
                    ")\n"
                )
            f.write("\n")

        # Write generalizations
        if model.generalizations:
            f.write("# Generalizations\n")
            for gen in sort(model.generalizations):
                specific_var_name = safe_class_name(gen.specific.name)
                general_var_name = safe_class_name(gen.general.name)
                f.write(f"gen_{gen.specific.name}_{gen.general.name} = Generalization"
                        f"(general={general_var_name}, specific={specific_var_name})\n")
            f.write("\n")

        # Write OCL constraints if they exist
        if hasattr(model, 'constraints') and model.constraints:
            f.write("\n# OCL Constraints\n")
            for constraint in sort(model.constraints):
                constraint_name = constraint.name.replace("-", "_")
                context_var_name = safe_class_name(constraint.context.name)
                f.write(f"{constraint_name}: Constraint = Constraint(\n")
                f.write(f"    name=\"{constraint.name}\",\n")
                f.write(f"    context={context_var_name},\n")
                f.write(f"    expression=\"{constraint.expression}\",\n")
                f.write(f"    language=\"{constraint.language}\"\n")
                f.write(")\n")
            f.write("\n")

        # Write domain model
        f.write("# Domain Model\n")
        f.write("domain_model = DomainModel(\n")
        f.write(f"    name=\"{model.name}\",\n")
        class_names = ', '.join(safe_class_name(cls.name) for cls in sort(model.get_classes()))
        enum_names = ', '.join(safe_class_name(enum.name) for enum in sort(model.get_enumerations()))
        types_str = (f"{class_names}, {enum_names}" if class_names and enum_names else
                    class_names or enum_names)
        f.write(f"    types={{{types_str}}},\n")
        if model.associations:
            f.write(f"    associations={{{', '.join(assoc.name for assoc in sort(model.associations))}}},\n")
        else:
            f.write("    associations={},\n")
        if hasattr(model, 'constraints') and model.constraints:
            constraints_str = ', '.join(c.name.replace("-", "_") for c in sort(model.constraints))
            f.write(f"    constraints={{{constraints_str}}},\n")
        if model.generalizations:
            f.write(f"    generalizations={{{', '.join(f'gen_{gen.specific.name}_{gen.general.name}' for gen in sort(model.generalizations))}}}\n")
        else:
            f.write("    generalizations={}\n")
        f.write(")\n")

    print(f"BUML model saved to {file_path}")


from besser.BUML.metamodel.ocl.ocl import *

class Factory:

    def __init__(self,context):
        self.context = context

    def checkInAttributes(self,name,context):
        for attrib in context.attributes:
            if name == attrib.name:
                return attrib
    def checkInAssociation(self,name,context):
        for end in context.all_association_ends():
            # for end in association.ends:
                if name == end.name:
                    return end

    def handleProp(self,name,iterator):
        if "." in name:
            name = name.split('.')[1]

        prop = self.checkInAttributes(name,self.context)
        if prop is None:
            prop = self.checkInAssociation(name,self.context)
        if iterator is not None and prop is None and len(iterator)>0:
            prop = self.checkInAttributes(name,iterator[-1].get_iterator[-1].type)
            if prop is None:
                prop = self.checkInAssociation(name, iterator[-1].get_iterator[-1].type)
            pass
        return prop

    def create_property_Call_Expression(self,name, type, iterators=None):
        prop = self.handleProp(name,iterators)
        if prop is not None:
            return prop
        else:
            raise Exception("Property "+name+ " not found in class "+str(self.context.name))

    def create_date_literal_expression(self,name,value):
        return DateLiteralExpression(name,value)
    def create_variable_expression(self,name,type):
        var = VariableExp(name,type)
        return var

    def create_boolean_literal_expression(self, name, val):
        return BooleanLiteralExpression(name, val)
    def create_integer_literal_expression(self, name, val):
        return IntegerLiteralExpression(name, val)
    def create_string_literal_expression(self,name,val):
        return StringLiteralExpression(name,val)

    def create_real_literal_expression(self, name, val):
        return RealLiteralExpression(name, val)
        pass

    def create_if_else_exp(self,name, type):
        return IfExp(name,type)
    def create_set_type(self):
        return SetType("Set")
    def create_ordered_set_type(self):
        return OrderedSetType("OrderedSetType")
    def create_sub_ordered_set_type(self):
        return OrderedSetType("SubOrderedSetType")

    def create_operation_call_expression(self, leftpart = None, rightpart= None, infixOperator= None, inBetweenOp=None,beforeOp = None,isleftNone=False,name = None):
        if inBetweenOp is None and isleftNone is False and name is None:
            if beforeOp is None:
                return OperationCallExpression("Operation", infixOperator.get_infix_operator(),
                                           [leftpart, infixOperator, rightpart])
            else:
                return OperationCallExpression("Operation", infixOperator.get_infix_operator(),
                                           [infixOperator, rightpart])
        elif inBetweenOp is None and isleftNone is True and name is None:
            return OperationCallExpression("Operation", infixOperator.get_infix_operator(),
                                           [infixOperator, rightpart])
        elif inBetweenOp is not None and name is None:
            if beforeOp is None:
                oce = OperationCallExpression("Operation", infixOperator.get_infix_operator(),
                                               [leftpart, infixOperator, rightpart])
                oce.referredOperation=inBetweenOp
                return oce
            else:
                oce = OperationCallExpression("Operation", infixOperator.get_infix_operator(),
                                              [beforeOp,leftpart, infixOperator, rightpart])
                oce.referredOperation=inBetweenOp
                return oce
        else:
            return OperationCallExpression(name=name, operation =name,arguments=[] )

    def create_type_exp(self,classifier):
        return TypeExp(classifier,classifier)

    def create_loop_expression(self,collectionOperator):
        return LoopExp(collectionOperator,None)
        pass
    def create_bag_type(self):
        return BagType("BagType")
    def create_collection_literal_expression(self,name,type):
        return CollectionLiteralExp(name = name,type=type)
    def create_sequence_type(self):
        return SequenceType("SequenceType")
    def create_sub_sequence_type(self):
        return SequenceType("SubSequenceType")
    def create_collection_item(self,name ,item):
        return CollectionItem(name,item)
    def create_iterator_expression(self,name,type):
        return IteratorExp(name,type)
        pass

    def create_infix_operator(self,op):
        return InfixOperator(op)



from enum import Enum
from besser.BUML.metamodel.structural import NamedElement, DomainModel, Model


class Hypervisor(Enum):
    '''Enumeration to list various types of hypervisors.

    Hypervisors are software, firmware, or hardware that creates and runs virtual machines (VMs). 
    This enumeration lists some of the most common hypervisors used in virtualization technology.

    Attributes:
        vm_ware (str): Represents the VMWare hypervisor.
        hyper_v (str): Represents the Hyper-V hypervisor developed by Microsoft.
        xen_server (str): Represents the XenServer hypervisor.
        rhev (str): Represents the Red Hat Enterprise Virtualization (RHEV) hypervisor.
        kvm (str): Represents the Kernel-based Virtual Machine (KVM) hypervisor.
    '''

    vm_ware = "VMWare"
    hyper_v = "Hyper-V"
    xen_server = "XenServer"
    rhev = "RHEV"
    kvm = "KVM"


class Processor(Enum):
    '''Enumeration to list various types of processors.

    Processors are the central units in computing that execute instructions.

    Attributes:
        x64 (str): Represents the x64 processor architecture.
        x86 (str): Represents the x86 processor architecture.
        arm (str): Represents the ARM processor architecture.
    '''

    x64 = "x64"
    x86 = "x84"
    arm = "ARM"


class IPRangeType(Enum):
    '''Enumeration to list different types of IP ranges.

    IP ranges are used to define segments of IP addresses in networking.

    Attributes:
        subnet (str): Represents a subnetwork IP range.
        pod (str): Represents a pod IP range.
        service (str): Represents a service IP range.
    '''

    subnet = "Subnetwork"
    pod = "Pod"
    service = "Service"


class ServiceType(Enum):
    '''Enumeration to list different types of services in networking.

    Services are abstractions that define a set of pods as a group and access policy for them.

    Attributes:
        lb (str): Represents a LoadBalancer service.
        ingress (str): Represents an Ingress service.
        egress (str): Represents an Egress service.
    '''

    lb = "LoadBalancer"
    ingress = "Ingress"
    egress = "Egress"


class Provider(Enum):
    '''Enumeration to list different cloud service providers.

    Cloud service providers offer various services such as computing power, storage, and networking.

    Attributes:
        google (str): Represents Google Cloud Platform.
        aws (str): Represents Amazon Web Services.
        azure (str): Represents Microsoft Azure.
        other (str): Represents any other cloud service provider.
    '''

    google = "Google"
    aws = "AWS"
    azure = "Azure"
    other = "Other"


class Protocol(Enum):
    '''Enumeration to list various network protocols.

    Network protocols define the rules for data communication over a network.

    Attributes:
        http (str): Represents the HTTP protocol.
        https (str): Represents the HTTPS protocol.
        tcp (str): Represents the TCP protocol.
        udp (str): Represents the UDP protocol.
        all (str): Represents all protocols.
    '''

    http = "HTTP"
    https = "HTTPS"
    tcp = "TCP"
    udp = "UPD"
    all = "ALL"


class Resources:
    """A class to represent the computational resources.

    Args:
        cpu (int): The number of CPU units.
        memory (int): The amount of memory in megabytes.

    Attributes:
        cpu (int): The number of CPU units.
        memory (int): The amount of memory in megabytes.
    """

    def __init__(self, cpu: int, memory: int):
        self.cpu: int = cpu
        self.memory: int = memory

    @property
    def cpu(self) -> int:
        """str: Get the cpu value."""
        return self.__cpu

    @cpu.setter
    def cpu(self, cpu: int):
        """str: Set the cpu value."""
        self.__cpu = cpu

    @property
    def memory(self) -> int:
        """str: Get the memory value."""
        return self.__memory

    @memory.setter
    def memory(self, memory: int):
        """str: Set the memory value."""
        self.__memory = memory

    def __repr__(self) -> str:
        return f'Resource({self.cpu}, {self.memory})'


class Application(NamedElement):
    """A class to represent an application.

    Args:
        name (str): The name of the application.
        image_repo (str): The image repository for the application.
        port (int): The port on which the application runs.
        required_resources (Resources): The resources required by the application.
        domain_model (DomainModel): The domain model of the application.

    Attributes:
        image_repo (str): The image repository for the application.
        port (int): The port on which the application runs.
        required_resources (Resources): The resources required by the application.
        domain_model (DomainModel): The domain model of the application.
    """

    def __init__(self, name: str, image_repo: str, port: int, required_resources: Resources,
            domain_model: DomainModel):
        super().__init__(name)
        self.image_repo: str = image_repo
        self.port: str = port
        self.required_resources: Resources = required_resources
        self.domain_model: DomainModel = domain_model

    @property
    def image_repo(self) -> str:
        """str: Get the image repository."""
        return self.__image_repo

    @image_repo.setter
    def image_repo(self, image_repo: str):
        """str: Set the image repository."""
        self.__image_repo = image_repo

    @property
    def port(self) -> int:
        """str: Get the application port."""
        return self.__port

    @port.setter
    def port(self, port: int):
        """str: Set the application port."""
        self.__port = port

    @property
    def required_resources(self) -> Resources:
        """Resource: Get the required resources."""
        return self.__required_resources

    @required_resources.setter
    def required_resources(self, required_resources: Resources):
        """Resource: Set the required resources."""
        self.__required_resources = required_resources

    @property
    def domain_model(self) -> DomainModel:
        """str: Get the domain model."""
        return self.__domain_model

    @domain_model.setter
    def domain_model(self, domain_model: DomainModel):
        """str: Set the domain model."""
        self.__domain_model = domain_model

    def __repr__(self) -> str:
        return (
            f'Application({self.name}, {self.required_resources}, {self.image_repo}, '
            f'{self.domain_model})'
        )


class Volume(NamedElement):
    """A class to represent a volume.

    Args:
        name (str): The name of the volume.
        mount_path (str): The mount path of the volume.
        sub_path (str): The sub-path within the volume.

    Attributes:
        mount_path (str): The mount path of the volume.
        sub_path (str): The sub-path within the volume.
    """

    def __init__(self, name: str, mount_path: str, sub_path: str):
        super().__init__(name)
        self.mount_path: str = mount_path
        self.sub_path: str = sub_path

    @property
    def mount_path(self) -> str:
        """str: Get the mount path."""
        return self.__mount_path

    @mount_path.setter
    def mount_path(self, mount_path: str):
        """str: Set the mount path."""
        self.__mount_path = mount_path

    @property
    def sub_path(self) -> str:
        """str: Get the sub path."""
        return self.__sub_path

    @sub_path.setter
    def sub_path(self, sub_path: str):
        """str: Set the sub path."""
        self.__sub_path = sub_path

    def __repr__(self) -> str:
        return f'Volume({self.name}, {self.mount_path}, {self.sub_path})'


class Container(NamedElement):
    """A class to represent a container.

    Args:
        name (str): The name of the container.
        application (Application): The application running in the container.
        resources_limit (Resources, optional): The resource limits for the container.
        volumes (set[Volume], optional): The set of volumes attached to the container.

    Attributes:
        application (Application): The application running in the container.
        resources_limit (Resources): The resource limits for the container.
        volumes (set[Volume]): The set of volumes attached to the container.
    """

    def __init__(self, name: str, application: Application, resources_limit: Resources = None,
            volumes: set[Volume] = None):
        super().__init__(name)
        self.application: Application = application
        self.resources_limit: Resources = resources_limit
        self.volumes: set[Volume] = volumes if volumes is not None else set()

    @property
    def application(self) -> Application:
        """Application: Get the application."""
        return self.__application

    @application.setter
    def application(self, application: Application):
        """Application: Set the application."""
        self.__application = application

    @property
    def resources_limit(self) -> Resources:
        """Resource: Get the resources limit."""
        return self.__resources_limit

    @resources_limit.setter
    def resources_limit(self, resources_limit: Resources):
        """Resource: Set the resources limit."""
        self.__resources_limit = resources_limit

    @property
    def volumes(self) -> set[Volume]:
        """set[Volume]: Get the set of volumes."""
        return self.__volumes

    @volumes.setter
    def volumes(self, volumes: set[Volume]):
        """set[Volume]: Set the set of volumes."""
        self.__volumes = volumes

    def __repr__(self) -> str:
        return f'Container({self.name}, {self.application}, {self.resources_limit}, {self.volumes})'


class Deployment(NamedElement):
    """ A class to represent a deployment.

    Args:
        name (str): The name of the deployment.
        replicas (int): The number of replicas.
        containers (set[Container]): The set of containers in the deployment.

    Attributes:
        replicas (int): The number of replicas.
        containers (set[Container]): The set of containers in the deployment.
    """

    def __init__(self, name: str, replicas: int, containers: set[Container]):
        super().__init__(name)
        self.replicas: int = replicas
        self.containers: set[Container] = containers

    @property
    def replicas(self) -> int:
        """int: Get the number of replicas."""
        return self.__replicas

    @replicas.setter
    def replicas(self, replicas: int):
        """int: Set the number of replicas."""
        self.__replicas = replicas

    @property
    def containers(self) -> set[Container]:
        """set[Container]: Get the set of containers."""
        return self.__containers

    @containers.setter
    def containers(self, containers: set[Container]):
        """set[Container]: Set the set of containers."""
        self.__containers = containers

    def __repr__(self) -> str:
        return f'Deployment({self.name}, {self.replicas}, {self.containers})'


class Service(NamedElement):
    """A class to represent a service.

    Args:
        name (str): The name of the service.
        port (int): The port on which the service is exposed.
        target_port (int): The port on which the application is running.
        type (ServiceType): The type of service (e.g., LoadBalancer, Ingress, Egress).
        protocol (Protocol): The protocol used by the service (e.g., HTTP, HTTPS, TCP, UDP).
        application (Application, optional): The application associated with the service.

    Attributes:
        port (int): The port on which the service is exposed.
        target_port (int): The port on which the application is running.
        type (ServiceType): The type of service.
        protocol (Protocol): The protocol used by the service.
        application (Application): The application associated with the service.
    """

    def __init__(self, name: str, port: int, target_port: int, type: ServiceType,
                 protocol: Protocol, application: Application = None):
        super().__init__(name)
        self.port: int = port
        self.target_port: int = target_port
        self.type: ServiceType = type
        self.protocol: Protocol = protocol
        self.application: Application = application

    @property
    def port(self) -> int:
        """int: Get the port on which the service is exposed."""
        return self.__port

    @port.setter
    def port(self, port: int):
        """int: Set the port on which the service is exposed."""
        self.__port = port

    @property
    def target_port(self) -> int:
        """int: Get the port on which the application is running."""
        return self.__target_port

    @target_port.setter
    def target_port(self, target_port: int):
        """int: Set the port on which the application is running."""
        self.__target_port = target_port

    @property
    def type(self) -> ServiceType:
        """ServiceType: Get the type of service."""
        return self.__type

    @type.setter
    def type(self, type: ServiceType):
        """ServiceType: Set the type of service."""
        self.__type = type

    @property
    def protocol(self) -> Protocol:
        """Protocol: Get the protocol used by the service."""
        return self.__protocol

    @protocol.setter
    def protocol(self, protocol: Protocol):
        """Protocol: Set the protocol used by the service."""
        self.__protocol = protocol

    @property
    def application(self) -> Application:
        """Application: Get the application associated with the service."""
        return self.__application

    @application.setter
    def application(self, application: Application):
        """Application: Set the application associated with the service."""
        self.__application = application

    def __repr__(self) -> str:
        return (
            f'Service({self.name}, {self.port}, {self.target_port}, {self.type}, '
            f'{self.protocol}, {self.application})'
        )

class IPRange(NamedElement):
    """A class to represent an IP range.

    Args:
        name (str): The name of the IP range.
        cidr_range (str): The CIDR range of the IP addresses.
        type (IPRangeType): The type of IP range (e.g., Subnetwork, Pod, Service).
        public (bool): Whether the IP range is public or private.

    Attributes:
        cidr_range (str): The CIDR range of the IP addresses.
        type (IPRangeType): The type of IP range.
        public (bool): Whether the IP range is public or private.
    """

    def __init__(self, name: str, cidr_range: str, type: IPRangeType, public: bool):
        super().__init__(name)
        self.cidr_range: str = cidr_range
        self.type: IPRangeType = type
        self.public: bool = public

    @property
    def cidr_range(self) -> str:
        """str: Get the CIDR range of the IP addresses."""
        return self.__cidr_range

    @cidr_range.setter
    def cidr_range(self, cidr_range: str):
        """str: Set the CIDR range of the IP addresses."""
        self.__cidr_range = cidr_range

    @property
    def type(self) -> IPRangeType:
        """IPRangeType: Get the type of IP range."""
        return self.__type

    @type.setter
    def type(self, type: IPRangeType):
        """IPRangeType: Set the type of IP range."""
        self.__type = type

    @property
    def public(self) -> bool:
        """bool: Get whether the IP range is public or private."""
        return self.__public

    @public.setter
    def public(self, public: bool):
        """bool: Set whether the IP range is public or private."""
        self.__public = public

    def __repr__(self) -> str:
        return f'IPRange({self.name}, {self.cidr_range}, {self.type}, {self.public})'


class SecurityGroup(NamedElement):
    """A class to represent a security group.

    Args:
        name (str): The name of the security group.
        rules (set[Service]): The set of services that define the security rules.

    Attributes:
        rules (set[Service]): The set of services that define the security rules.
    """

    def __init__(self, name: str, rules: set[Service]):
        super().__init__(name)
        self.rules: set[Service] = rules

    @property
    def rules(self) -> set[Service]:
        """set[Service]: Get the set of security rules."""
        return self.__rules

    @rules.setter
    def rules(self, rules: set[Service]):
        """set[Service]: Set the security rules."""
        self.__rules = rules

    def __repr__(self) -> str:
        return f'SecurityGroup({self.name}, {self.rules})'


class Network(NamedElement):
    """A class to represent a network.

    Args:
        name (str): The name of the network.
        security_groups (set[SecurityGroup], optional): The set of security groups associated
        with the network.

    Attributes:
        security_groups (set[SecurityGroup]): The set of security groups associated with the
        network.
    """

    def __init__(self, name: str, security_groups: set[SecurityGroup] = None):
        super().__init__(name)
        self.security_groups: set[SecurityGroup] = security_groups if security_groups is not None else set()

    @property
    def security_groups(self) -> set[SecurityGroup]:
        """set[SecurityGroup]: Get the set of security groups."""
        return self.__security_groups

    @security_groups.setter
    def security_groups(self, security_groups: set[SecurityGroup]):
        """set[SecurityGroup]: Set the security groups."""
        self.__security_groups = security_groups

    def __repr__(self) -> str:
        return f'Network({self.name})'


class Subnetwork(NamedElement):
    """A class to represent a subnetwork.

    Args:
        name (str): The name of the subnetwork.
        ip_ranges (set[IPRange]): The set of IP ranges within the subnetwork.
        network (Network): The network to which the subnetwork belongs.

    Attributes:
        ip_ranges (set[IPRange]): The set of IP ranges within the subnetwork.
        network (Network): The network to which the subnetwork belongs.
    """

    def __init__(self, name: str, ip_ranges: set[IPRange], network: Network):
        super().__init__(name)
        self.ip_ranges : set[IPRange] = ip_ranges
        self.network: Network = network

    @property
    def ip_ranges(self) -> set[IPRange]:
        """set[IPRange]: Get the set of IP ranges."""
        return self.__ip_ranges

    @ip_ranges.setter
    def ip_ranges(self, ip_ranges: set[IPRange]):
        """set[IPRange]: Set the set of IP ranges."""
        self.__ip_ranges = ip_ranges

    @property
    def network(self) -> Network:
        """Network: Get the network."""
        return self.__network

    @network.setter
    def network(self, network: Network):
        """Network: Set the network."""
        self.__network = network

    def __repr__(self) -> str:
        return f'Subnetwork({self.name}, {self.ip_ranges}, {self.network})'


class Zone(NamedElement):
    """A class to represent a zone.

    Args:
        name (str): The name of the zone.

    Attributes:
        name (str): The name of the zone.
    """

    def __init__(self, name: str):
        super().__init__(name)

    def __repr__(self) -> str:
        return f'Zone({self.name})'

class Region(NamedElement):
    """A class to represent a region.

    Args:
        name (str): The name of the region.
        zones (set[Zone]): The set of zones within the region.

    Attributes:
        zones (set[Zone]): The set of zones within the region.
    """

    def __init__(self, name: str, zones: set[Zone]):
        super().__init__(name)
        self.zones : set[Zone] = zones

    @property
    def zones(self) -> set[Zone]:
        """set[Zone]: Get the set of zones."""
        return self.__zones

    @zones.setter
    def zones(self, zones: set[Zone]):
        """set[Zone]: Set the set of zones."""
        self.__zones = zones

    def __repr__(self) -> str:
        return f'Region({self.name}, {self.zones})'


class Node(NamedElement):
    """A class to represent a node.

    Args:
        name (str): The name of the node.
        public_ip (str): The public IP address of the node.
        private_ip (str): The private IP address of the node.
        os (str): The operating system running on the node.
        resources (Resources): The computational resources of the node.
        storage (int): The storage capacity of the node in gigabytes.
        processor (Processor): The processor type of the node.

    Attributes:
        public_ip (str): The public IP address of the node.
        private_ip (str): The private IP address of the node.
        os (str): The operating system running on the node.
        resources (Resources): The computational resources of the node.
        storage (int): The storage capacity of the node in gigabytes.
        processor (Processor): The processor type of the node.
    """

    def __init__(self, name: str, public_ip: str, private_ip, os: str, resources: Resources,
                 storage: int, processor: Processor):
        super().__init__(name)
        self.public_ip: str = public_ip
        self.private_ip: str = private_ip
        self.os: str = os
        self.resources: Resources = resources
        self.storage: int = storage
        self.processor: Processor = processor

    @property
    def public_ip(self) -> str:
        """str: Get the public IP address of the node."""
        return self.__public_ip

    @public_ip.setter
    def public_ip(self, public_ip: str):
        """str: Set the public IP address of the node."""
        self.__public_ip = public_ip

    @property
    def private_ip(self) -> str:
        """str: Get the private IP address of the node."""
        return self.__private_ip

    @private_ip.setter
    def private_ip(self, private_ip: str):
        """str: Set the private IP address of the node."""
        self.__private_ip = private_ip

    @property
    def os(self) -> str:
        """str: Get the operating system running on the node."""
        return self.__os

    @os.setter
    def os(self, os: str):
        """str: Set the operating system running on the node."""
        self.__os = os

    @property
    def resources(self) -> Resources:
        """Resources: Get the computational resources of the node."""
        return self.__resources

    @resources.setter
    def resources(self, resources: Resources):
        """Resources: Set the computational resources of the node."""
        self.__resources = resources

    @property
    def storage(self) -> int:
        """int: Get the storage capacity of the node in gigabytes."""
        return self.__storage

    @storage.setter
    def storage(self, storage: int):
        """int: Set the storage capacity of the node in gigabytes."""
        self.__storage = storage

    @property
    def processor(self) -> Processor:
        """Processor: Get the processor type of the node."""
        return self.__processor

    @processor.setter
    def processor(self, processor: Processor):
        """Processor: Set the processor type of the node."""
        self.__processor = processor


class EdgeNode(Node):
    """A class to represent an edge node.
    """

    def __init__(self, name: str, public_ip: str, private_ip, os: str, resources: Resources,
                 storage: int, processor: Processor):
        super().__init__(name, public_ip, private_ip, os, resources, storage, processor)


class CloudNode(Node):
    """A class to represent a cloud node.
    """

    def __init__(self, name: str, public_ip: str, private_ip, os: str, resources: Resources,
                 storage: int, processor: Processor):
        super().__init__(name, public_ip, private_ip, os, resources, storage, processor)


class Cluster(NamedElement):
    """ A class to represent a cluster.

    Args:
        name (str): The name of the cluster.
        services (set[Service]): The set of services associated with the cluster.
        deployments (set[Deployment]): The set of deployments in the cluster.
        regions (set[Region]): The set of regions where the cluster is deployed.
        net_config (bool, optional): Indicates if network configuration is enabled. Defaults to True.
        nodes (set[Node], optional): The set of nodes in the cluster. Defaults to an empty set.
        networks (set[Network], optional): The set of networks in the cluster. Defaults to an empty set.
        subnets (set[Subnetwork], optional): The set of subnetworks in the cluster. Defaults to an empty set.

    Attributes:
        services (set[Service]): The set of services associated with the cluster.
        deployments (set[Deployment]): The set of deployments in the cluster.
        regions (set[Region]): The set of regions where the cluster is deployed.
        net_config (bool): Indicates if network configuration is enabled.
        nodes (set[Node]): The set of nodes in the cluster.
        networks (set[Network]): The set of networks in the cluster.
        subnets (set[Subnetwork]): The set of subnetworks in the cluster.
    """

    def __init__(self, name: str, services: set[Service], deployments: set[Deployment],
                 regions: set[Region], net_config: bool = True, nodes: set[Node] = None,
                 networks: set[Network] = None, subnets: set[Subnetwork] = None):
        super().__init__(name)
        self.services: set[Service] = services
        self.deployments: set[Deployment] = deployments
        self.regions: set[Region] = regions
        self.net_config: bool = net_config
        self.nodes: set[Node] = nodes
        self.networks: set[Network] = networks
        self.subnets: set[Subnetwork] = subnets

    @property
    def services(self) -> set[Service]:
        """set[Service]: Get the set of services."""
        return self.__services

    @services.setter
    def services(self, services: set[Service]):
        """set[Service]: Set the set of services."""
        self.__services = services

    @property
    def deployments(self) -> set[Deployment]:
        """set[Deployment]: Get the set of deployments."""
        return self.__deployments

    @deployments.setter
    def deployments(self, deployments: set[Deployment]):
        """set[Deployment]: Set the set of deployments."""
        self.__deployments = deployments

    @property
    def regions(self) -> set[Region]:
        """set[Region]: Get the set of regions."""
        return self.__regions

    @regions.setter
    def regions(self, regions: set[Region]):
        """set[Region]: Set the set of regions."""
        self.__regions = regions

    @property
    def net_config(self) -> bool:
        """bool: Get the network configuration status."""
        return self.__net_config

    @net_config.setter
    def net_config(self, net_config: bool):
        """bool: Set the network configuration status."""
        self.__net_config = net_config

    @property
    def nodes(self) -> set[Node]:
        """set[Node]: Get the set of nodes."""
        return self.__nodes

    @nodes.setter
    def nodes(self, nodes: set[Node]):
        """set[Node]: Set the set of nodes."""
        self.__nodes = nodes

    @property
    def networks(self) -> set[Network]:
        """set[Network]: Get the set of networks."""
        return self.__networks

    @networks.setter
    def networks(self, networks: set[Network]):
        """set[Network]: Set the set of networks."""
        self.__networks = networks

    @property
    def subnets(self) -> set[Subnetwork]:
        """set[Subnetwork]: Get the set of subnetworks."""
        return self.__subnets

    @subnets.setter
    def subnets(self, subnets: set[Subnetwork]):
        """set[Subnetwork]: Set the set of subnetworks."""
        self.__subnets = subnets

    def __repr__(self) -> str:
        return (
            f'Cluster({self.name}, {self.services}, {self.deployments}, {self.regions}, '
            f'{self.net_config}, {self.nodes}, {self.networks}, {self.subnets})'
        )


class PublicCluster(Cluster):
    """A class to represent a public cluster.

    Args:
        name (str): The name of the public cluster.
        services (set[Service]): The set of services associated with the public cluster.
        deployments (set[Deployment]): The set of deployments in the public cluster.
        regions (set[Region]): The set of regions where the public cluster is deployed.
        num_nodes (int): The number of nodes in the public cluster.
        provider (Provider): The provider of the public cluster.
        config_file (str): The configuration file for the public cluster.
        networks (set[Network], optional): The set of networks in the public cluster. Defaults to an empty set.
        subnets (set[Subnetwork], optional): The set of subnetworks in the public cluster. Defaults to an empty set.
        net_config (bool, optional): Indicates if network configuration is enabled. Defaults to True.

    Attributes:
        config_file (str): The configuration file for the public cluster.
        num_nodes (int): The number of nodes in the public cluster.
        provider (Provider): The provider of the public cluster.
    """

    def __init__(self, name: str, services: set[Service], deployments: set[Deployment],
                 regions: Region, num_nodes: int, provider: Provider, config_file: str,
                 networks: set[Network] = None, subnets: set[Subnetwork] = None,
                 net_config: bool = True):
        super().__init__(name, services, deployments, regions, net_config, networks, subnets)
        self.config_file: str = config_file
        self.num_nodes: int = num_nodes
        self.provider: Provider = provider

    @property
    def config_file(self) -> str:
        """str: Get the configuration file."""
        return self.__config_file

    @config_file.setter
    def config_file(self, config_file: str):
        """str: Set the configuration file."""
        self.__config_file = config_file

    @property
    def num_nodes(self) -> int:
        """int: Get the number of nodes."""
        return self.__num_nodes

    @num_nodes.setter
    def num_nodes(self, num_nodes: int):
        """int: Set the number of nodes."""
        self.__num_nodes = num_nodes

    @property
    def provider(self) -> Provider:
        """Provider: Get the provider."""
        return self.__provider

    @provider.setter
    def provider(self, provider: Provider):
        """Provider: Set the provider."""
        self.__provider = provider

    def __repr__(self) -> str:
        return (
            f'PublicCluster({self.name}, {self.services},{self.deployments}, '
            f'{self.regions}, {self.config_file}, {self.provider}, '
            f'{self.networks}, {self.subnets}, {self.net_config})'
        )

class OnPremises(Cluster):
    """A class to represent an on-premises cluster.

    Args:
        name (str): The name of the on-premises cluster.
        services (set[Service]): The set of services associated with the on-premises cluster.
        deployments (set[Deployment]): The set of deployments in the on-premises cluster.
        regions (set[Region]): The set of regions where the on-premises cluster is deployed.
        nodes (set[Node]): The set of nodes in the on-premises cluster.
        hypervisor (Hypervisor): The hypervisor used in the on-premises cluster.
        networks (set[Network]): The set of networks in the on-premises cluster.
        subnets (set[Subnetwork]): The set of subnetworks in the on-premises cluster.

    Attributes:
        hypervisor (Hypervisor): The hypervisor used in the on-premises cluster.
    """

    def __init__(self, name: str, services: set[Service], deployments: set[Deployment],
                 regions: Region, nodes: set[Node], hypervisor: Hypervisor,
                 networks: set[Network], subnets: set[Subnetwork]):
        super().__init__(name, services, deployments, regions, nodes, networks, subnets)
        self.hypervisor: str = hypervisor

    @property
    def hypervisor(self) -> Hypervisor:
        """Hypervisor: Get the hypervisor."""
        return self.__hypervisor

    @hypervisor.setter
    def hypervisor(self, hypervisor: Hypervisor):
        """Hypervisor: Set the hypervisor."""
        self.__hypervisor = hypervisor

    def __repr__(self) -> str:
        return (
            f'Cluster({self.name}, {self.services},{self.deployments}, {self.regions}, '
            f'{self.nodes}, {self.hypervisor}, {self.networks}, {self.subnets})'
        )


class DeploymentModel(Model):
    """A class to represent a deployment model.

    Args:
        name (str): The name of the deployment model.
        clusters (set[Cluster]): The set of clusters in the deployment model.

    Attributes:
        clusters (set[Cluster]): The set of clusters in the deployment model.
    """

    def __init__(self, name: str, clusters: set[Cluster]):
        super().__init__(name)
        self.clusters: set[Cluster] = clusters

    @property
    def clusters(self) -> set[Cluster]:
        """set[Cluster]: Get the set of clusters."""
        return self.__clusters

    @clusters.setter
    def clusters(self, clusters: set[Cluster]):
        """set[Cluster]: Set the set of clusters."""
        self.__clusters = clusters

    def __repr__(self) -> str:
        return f'Cluster({self.name}, {self.clusters})'



import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.deployment import DeploymentModel, IPRangeType, ServiceType, Protocol
from besser.generators import GeneratorInterface

class TerraformGenerator(GeneratorInterface):
    """
    TerraformGenerator is a class that implements the GeneratorInterface and is responsible for generating
    the Terraform models code based on the input B-UML model.
    Args:
        deployement_model (DeployementModel): The deployment model containing multiple public clusters.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    def __init__(self, deployment_model: DeploymentModel, output_dir: str = None):
        super().__init__(deployment_model, output_dir)
        self.deployment_model = deployment_model
        os.makedirs(self.output_dir, exist_ok=True)
        self.env = self.setup_environment()

    def setup_environment(self):
        """
        Sets up the Jinja2 environment and adds custom filters.
        """
        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path), trim_blocks=True, lstrip_blocks=True)
        env.filters['class_name'] = self.get_class_name
        env.filters['to_str'] = self.convert_type
        env.filters['set_dict_item'] = self.set_dict_item
        return env

    @staticmethod
    def get_class_name(value):
        """
        Custom filter to get the class name of an object.
        """
        return value.__class__.__name__

    @staticmethod
    def set_dict_item(dictionary, key, value):
        dictionary[key] = f"default-{value}"
        return ''

    @staticmethod
    def convert_type(value):
        """
        Converts the input value to its underlying value if it is an instance of an enum class,
        otherwise converts it to a string.
        Args:
            value: The value to convert, which can be of any type including enum instances.
        Returns:
            The underlying value if an enum instance, or the string representation of the value.
        """
        if isinstance(value, (IPRangeType, ServiceType, Protocol)):
            return value.value
        return str(value)

    def generate(self):
        """
        Generates Terraform code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory is not specified, the generated code will be stored in the <current directory>/output
        folder.

        Returns:
            None, but stores the generated code as files with specified names.
        """
        output_base_dir = self.build_generation_dir()

        for public_cluster in self.deployment_model.clusters:
            try:
                # Read the configuration file .conf line by line
                with open(public_cluster.config_file, 'r') as file:
                    config_lines = file.readlines()
            except IOError as e:
                print(f"Error reading configuration file {public_cluster.config_file}: {e}")
                continue

            # Dictionary mapping template file names to output file names
            template_to_file_map = self.get_template_to_file_map(public_cluster.provider.value)

            if not template_to_file_map:
                print(f"Unsupported provider: {public_cluster.provider.value}")
                continue

            # Create a directory path with the provider and cluster name
            cluster_dir = os.path.join(output_base_dir,
                                       f"{public_cluster.provider.value.lower()}_{public_cluster.name}")
            os.makedirs(cluster_dir, exist_ok=True)

            for template_name, output_file_name in template_to_file_map.items():
                file_path = os.path.join(cluster_dir, output_file_name)
                try:
                    template = self.env.get_template(template_name)
                except Exception as e:
                    print(f"Error loading template {template_name}: {e}")
                    continue

                try:
                    with open(file_path, mode="w") as f:
                        generated_code = template.render(public_cluster=public_cluster, config_lines=config_lines)
                        f.write(generated_code)
                        print(f"Code generated in the location: {file_path}")
                except IOError as e:
                    print(f"Error writing to file {file_path}: {e}")

    @staticmethod
    def get_template_to_file_map(provider):
        """
        Returns the template to file map based on the provider.

        Args:
            provider (str): The cloud provider (e.g., 'Google', 'AWS').

        Returns:
            dict: Mapping of template file names to output file names.
        """
        if provider == 'Google':
            return {
                'gcp/version.tf.j2': 'version.tf',
                'gcp/cluster.tf.j2': 'cluster.tf',
                'gcp/app.tf.j2': 'app.tf',
                'gcp/api.tf.j2': 'api.tf',
                'gcp/setup.bat.j2': 'setup.bat'
            }
        elif provider == 'AWS':
            return {
                'aws/eks.tf.j2': 'eks.tf',
                'aws/iam-oidc.tf.j2': 'iam-oidc.tf',
                'aws/provider.tf.j2': 'provider.tf',
                'aws/igw.tf.j2': 'igw.tf',
                'aws/nat.tf.j2': 'nat.tf',
                'aws/routes.tf.j2': 'routes.tf',
                'aws/vpc.tf.j2': 'vpc.tf',
                'aws/nodes.tf.j2': 'nodes.tf',
                'aws/subnets.tf.j2': 'subnets.tf',
                'aws/setup.bat.j2': 'setup.bat'
            }
        else:
            return None



"""
BUML code for a regression prediction problem.
The task is to predict house prices using the BostonHousing dataset.
"""

from besser.BUML.metamodel.nn import NN, LinearLayer, DropoutLayer, \
    Configuration, Dataset
from besser.generators.nn.pytorch.pytorch_code_generator import (
    PytorchGenerator
)
from besser.generators.nn.tf.tf_code_generator import TFGenerator

nn_model: NN = NN(name="my_model")
nn_model.add_layer(LinearLayer(name="l1", actv_func="relu", in_features=13,
                               out_features=64))
nn_model.add_layer(LinearLayer(name="l2", actv_func="relu", in_features=64,
                               out_features=128))
nn_model.add_layer(DropoutLayer(name="l3", rate=0.2))
nn_model.add_layer(LinearLayer(name="l4", in_features=128, out_features=1))

parameters: Configuration = Configuration(batch_size=6, epochs=40,
                                    learning_rate=0.001, optimizer="adam",
                                    loss_function="mse", metrics=["mae"])

nn_model.add_configuration(parameters)

train_data = Dataset(name="train_data",
                     path_data="dataset/BostonHousingTrain.csv",
                     task_type="regression", input_format="csv")
test_data = Dataset(name="test_data",
                    path_data="dataset/BostonHousingTest.csv")

nn_model.add_train_data(train_data)
nn_model.add_test_data(test_data)

pytorch_model = PytorchGenerator(model=nn_model,
                                 output_dir="output/regression")
pytorch_model.generate()

tf_model = TFGenerator(model=nn_model, output_dir="output/regression")
tf_model.generate()



import subprocess
import os
import sys
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.generators.sql_alchemy import SQLAlchemyGenerator

class SQLGenerator(GeneratorInterface):
    """
    SQLGenerator is a class that implements the GeneratorInterface and produces the code or set of SQL statements 
    used to define and modify the structure of the tables in a database.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
        sql_dialect (str, optional): The SQL dialect. Values allowed: "sqlite", "postgresql", "mysql",
        "mssql", or "mariadb". Defaults to "sqlite".
    """

    def __init__(self, model: DomainModel, output_dir: str = None, sql_dialect: str = "sqlite"):
        super().__init__(model, output_dir)
        self.sql_dialect = sql_dialect

    def generate(self):
        """
        Generates SQL code based on the provided B-UML model and saves it to the specified output directory.
        If the output directory was not specified, the code generated will be stored in the <current directory>/output
        folder.

        Returns:
            None, but store the generated code as a file named tables.sql.
        """
        file_path = self.build_generation_path(file_name="sql_alchemy.py")
        alchemy_gen = SQLAlchemyGenerator(model=self.model, output_dir=self.output_dir)
        alchemy_gen.generate(dbms=self.sql_dialect)
        with open(file_path, "r", encoding="utf-8") as file:
            content = file.read()
        # remove the last line
        lines = content.splitlines()
        lines = lines[:-1]
        content = "\n".join(lines)
        dialect = self.sql_dialect
        content += f"""

from sqlalchemy.schema import CreateTable
import os

current_directory = os.path.dirname(os.path.realpath(__file__))
sql_file_path = os.path.join(current_directory, f"tables_{dialect}.sql")

ddl_statements = []

for table in Base.metadata.sorted_tables:
    ddl_statements.append(str(CreateTable(table).compile(engine)))

with open(sql_file_path, "w") as f:
    for stmt in ddl_statements:
        f.write(f"{{stmt}};\\n\\n")
"""
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(content)

        # Execute the generated file
        try:
            #subprocess.run(['python', file_path], check=True, text=True, capture_output=True)
            subprocess.run([sys.executable, file_path], check=True, text=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            print(f"Error executing the sql_alchemy file: {e.stderr}")

        # Remove sql_alchemy.py
        try:
            os.remove(file_path)
        except Exception as e:
            print(f"Error removing the sql_alchemy file: {e}")



from tests.BUML.metamodel.object.library_object import library_model, object_model
from besser.BUML.notations.ocl.OCLParserWrapper import OCLParserWrapper

def test1():
    parser = OCLParserWrapper(library_model, object_model)
    cons = list(library_model.constraints)
    constraint = cons[0]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test2():
    parser = OCLParserWrapper(library_model, object_model)
    cons = list(library_model.constraints)
    constraint = cons[1]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test3():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[2]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test4():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[3]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test5():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[4]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test6():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[4]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test7():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[5]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test8():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[6]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None

    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test9():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[7]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test10():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[8]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test11():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[9]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

def test12():
    parser = OCLParserWrapper(library_model, object_model)

    cons = list(library_model.constraints)
    constraint = cons[10]
    print("Query: " + str(constraint.expression),end = ": ")
    res = None
    try:
        res = parser.parse(constraint)
    except Exception as error:
            print('\x1b[0;30;41m' + 'Exception Occured! Info:' +str(error)  + '\x1b[0m')
            res = None
    if res:
            print('\x1b[6;30;42m' + 'Parsed Correctly' + '\x1b[0m')
    elif res == False:
            print('\x1b[0;30;41m' + 'Cannot be Parsed' + '\x1b[0m')
    assert res == True

if __name__ == '__main__':
    pass